import pytest

from aesqlapius.args import prepare_args_as_dict, prepare_args_as_list
from aesqlapius.function_def import parse_function_definition


@pytest.fixture
def func_def():
    func_code = """def foo(a: str,
                b: int,
                c: str="c",
                d: int=4,
                e: str="e",
                f: int=6) -> None: ...
    """

    return parse_function_definition(func_code)


def test_dict_as_dict(func_def):
    assert prepare_args_as_dict(
        func_def,
        [
            'a',
            2,
            'cc',
            44
        ],
        {}
    ) == {
        'a': 'a',
        'b': 2,
        'c': 'cc',
        'd': 44,
        'e': 'e',
        'f': 6
    }


def test_list_as_dict(func_def):
    assert prepare_args_as_dict(
        func_def,
        [],
        {
            'a': 'a',
            'b': 2,
            'c': 'cc',
            'd': 44
        }
    ) == {
        'a': 'a',
        'b': 2,
        'c': 'cc',
        'd': 44,
        'e': 'e',
        'f': 6
    }


def test_dict_as_list(func_def):
    assert prepare_args_as_list(
        func_def,
        [
            'a',
            2,
            'cc',
            44
        ],
        {}
    ) == [
        'a',
        2,
        'cc',
        44,
        'e',
        6
    ]


def test_list_as_list(func_def):
    assert prepare_args_as_list(
        func_def,
        [],
        {
            'a': 'a',
            'b': 2,
            'c': 'cc',
            'd': 44
        }
    ) == [
        'a',
        2,
        'cc',
        44,
        'e',
        6
    ]


def test_missing_args(func_def):
    with pytest.raises(TypeError):
        prepare_args_as_list(func_def, [], {})


def test_duplicate_args(func_def):
    with pytest.raises(TypeError):
        prepare_args_as_list(func_def, ['a'], {'a': 'a'})



from dataclasses import dataclass

import pytest

from aesqlapius.function_def import ReturnValueInnerFormat
from aesqlapius.output import generate_row_processor


@pytest.fixture
def row():
    return (1, 2, 3)


@pytest.fixture
def field_names():
    return ['a', 'b', 'c']


def test_tuple(row, field_names):
    rp = generate_row_processor(ReturnValueInnerFormat.TUPLE, field_names)

    assert rp(row) == (1, 2, 3)


def test_dict(row, field_names):
    rp = generate_row_processor(ReturnValueInnerFormat.DICT, field_names)

    assert rp(row) == {'a': 1, 'b': 2, 'c': 3}


def test_list(row, field_names):
    rp = generate_row_processor(ReturnValueInnerFormat.LIST, field_names)

    assert rp(row) == [1, 2, 3]


def test_custom_local(row, field_names):
    @dataclass
    class MyLocalType:
        a: int
        b: int
        c: int

    rp = generate_row_processor('MyLocalType', field_names, 1)

    assert rp(row) == MyLocalType(1, 2, 3)


@dataclass
class MyGlobalType:
    a: int
    b: int
    c: int


def test_custom_global(row, field_names):
    rp = generate_row_processor('MyGlobalType', field_names, 1)

    assert rp(row) == MyGlobalType(1, 2, 3)


def test_custom_not_found(row, field_names):
    rp = generate_row_processor('MyNonexistingType', field_names, 1)

    with pytest.raises(NameError):
        rp(row)



import os
from typing import Iterator

import pytest

from aesqlapius.apis.psycopg2 import generate_api

psycopg2 = pytest.importorskip('psycopg2')
pytest_datadir = pytest.importorskip('pytest_datadir')


pytestmark = pytest.mark.skipif(
    not os.environ.get('POSTGRESQL_DSN'),
    reason='no POSTGRESQL_DSN defined, skipping PostgreSQL tests'
)


def test_generate_default(original_datadir):
    db = psycopg2.connect(os.environ.get('POSTGRESQL_DSN'))
    api = generate_api(original_datadir, db)

    assert api.ping() == {'pong': True}


def test_generate_nobind(original_datadir):
    api = generate_api(original_datadir)

    db = psycopg2.connect(os.environ.get('POSTGRESQL_DSN'))

    assert api.ping(db) == {'pong': True}


def test_generate_target(original_datadir):
    class MyDB():
        def __init__(self, dsn: str) -> None:
            self.db = psycopg2.connect(dsn)
            generate_api(original_datadir, self.db, target=self)

    db = MyDB(os.environ.get('POSTGRESQL_DSN'))

    assert db.ping() == {'pong': True}


def test_generate_ns_default(original_datadir):
    db = psycopg2.connect(os.environ.get('POSTGRESQL_DSN'))
    api = generate_api(original_datadir / 'sub', db)

    assert api.a() == ('a',)
    assert api.b() == ('b',)


def test_generate_ns_file(original_datadir):
    db = psycopg2.connect(os.environ.get('POSTGRESQL_DSN'))
    api = generate_api(original_datadir / 'sub', db, file_as_namespace=True)

    assert api.a.a() == ('a',)
    assert api.b.b() == ('b',)


def test_generate_ns_file_altroot(original_datadir):
    db = psycopg2.connect(os.environ.get('POSTGRESQL_DSN'))
    api = generate_api(original_datadir / 'sub', db, file_as_namespace=True, namespace_root='a')

    assert api.a() == ('a',)
    assert api.b.b() == ('b',)


@pytest.fixture()
def api(original_datadir):
    db = psycopg2.connect(os.environ.get('POSTGRESQL_DSN'))
    return generate_api(original_datadir, db)


def test_pass_args(api):
    assert(api.swap_args() == ('a', 0))
    assert(api.swap_args(1) == ('a', 1))
    assert(api.swap_args(a=1) == ('a', 1))
    assert(api.swap_args(b='b') == ('b', 0))
    assert(api.swap_args(1, 'b') == ('b', 1))
    assert(api.swap_args(a=1, b='b') == ('b', 1))
    assert(api.swap_args(b='b', a=1) == ('b', 1))


def test_get_nothing(api):
    api.get.nothing()


def test_get_iterator_tuple(api):
    it = api.get.iterator_tuple()
    assert(isinstance(it, Iterator))
    assert(next(it) == (0, 'a'))
    assert(next(it) == (1, 'b'))
    assert(next(it) == (2, 'c'))


def test_get_iterator_dict(api):
    it = api.get.iterator_dict()
    assert(isinstance(it, Iterator))
    assert(next(it) == {'a': 0, 'b': 'a'})
    assert(next(it) == {'a': 1, 'b': 'b'})
    assert(next(it) == {'a': 2, 'b': 'c'})


def test_get_iterator_list(api):
    it = api.get.iterator_list()
    assert(isinstance(it, Iterator))
    assert(next(it) == [0, 'a'])
    assert(next(it) == [1, 'b'])
    assert(next(it) == [2, 'c'])


def test_get_list_tuple(api):
    assert api.get.list_tuple() == [
        (0, 'a'),
        (1, 'b'),
        (2, 'c'),
    ]


def test_get_list_dict(api):
    assert api.get.list_dict() == [
        {'a': 0, 'b': 'a'},
        {'a': 1, 'b': 'b'},
        {'a': 2, 'b': 'c'},
    ]


def test_get_list_list(api):
    assert api.get.list_list() == [
        [0, 'a'],
        [1, 'b'],
        [2, 'c'],
    ]


def test_get_single_tuple(api):
    assert api.get.single_tuple() == (0, 'a')


def test_get_single_dict(api):
    assert api.get.single_dict() == {'a': 0, 'b': 'a'}


def test_get_single_list(api):
    assert api.get.single_list() == [0, 'a']



# Copyright (c) 2020 Dmitry Marakasov <amdmi3@amdmi3.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import inspect
from typing import Any, Callable, Dict, List, Tuple, Union

from aesqlapius.function_def import ReturnValueInnerFormat


def generate_row_processor(inner_format: Union[ReturnValueInnerFormat, str], field_names: List[str], stack_depth: int = 2) -> Callable[[Tuple[Any]], Any]:
    if isinstance(inner_format, str):
        def process_row_custom(row: Tuple[Any]) -> Dict[str, Any]:
            # get the frame of method caller; usually,
            # [0] is process_row
            # [1] is generated method
            # [2] is caller
            # but stack depth is customizable to simplify testing the function
            caller_frame = inspect.stack()[stack_depth].frame

            assert isinstance(inner_format, str)  # XXX: for mypy

            if inner_format in caller_frame.f_locals:
                custom_format = caller_frame.f_locals[inner_format]
            elif inner_format in caller_frame.f_globals:
                custom_format = caller_frame.f_globals[inner_format]
            else:
                raise NameError(f"name '{inner_format}' is not defined")

            return custom_format(**dict(zip(field_names, row)))

        return process_row_custom

    elif inner_format == ReturnValueInnerFormat.TUPLE:
        def process_row_tuple(row: Tuple[Any]) -> Tuple[Any]:
            return row

        return process_row_tuple

    elif inner_format == ReturnValueInnerFormat.DICT:
        def process_row_dict(row: Tuple[Any]) -> Dict[str, Any]:
            return dict(zip(field_names, row))

        return process_row_dict

    elif inner_format == ReturnValueInnerFormat.LIST:
        def process_row_list(row: Tuple[Any]) -> List[Any]:
            return list(row)

        return process_row_list

    else:
        raise NotImplementedError(f"unsupported inner return type format '{inner_format}'")  # pragma: no cover



# Copyright (c) 2020 Dmitry Marakasov <amdmi3@amdmi3.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dataclasses import dataclass
from typing import List, Optional

from aesqlapius.function_def import (
    FunctionDefinition,
    parse_function_definition
)


@dataclass
class Query:
    func_def: FunctionDefinition
    text: str


def parse_query_file(path: str) -> List[Query]:
    func_def: Optional[FunctionDefinition] = None
    text = ''
    res = []

    def flush():
        nonlocal func_def, text, res
        if func_def and text:
            res.append(Query(func_def, text))
        func_def = None
        text = ''

    with open(path, 'r') as fd:
        for line in fd:
            if line.startswith('-- def '):
                flush()
                func_def = parse_function_definition(line.removeprefix('-- '))

            text += line

    flush()

    return res



import pytest

from aesqlapius.function_def import (
    ArgumentDefinition,
    FunctionDefinition,
    ReturnValueDefinition,
    ReturnValueInnerFormat,
    ReturnValueOuterFormat,
    parse_function_definition
)


def test_simple():
    assert parse_function_definition(
        'def Foo() -> None: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=None
    )


def test_args_unannotated():
    assert parse_function_definition(
        'def Foo(a, b, c) -> None: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[
            ArgumentDefinition(name='a'),
            ArgumentDefinition(name='b'),
            ArgumentDefinition(name='c'),
        ],
        returns=None
    )


def test_args():
    assert parse_function_definition(
        'def Foo(a: int, b: str, c: float) -> None: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[
            ArgumentDefinition(name='a', type_='int'),
            ArgumentDefinition(name='b', type_='str'),
            ArgumentDefinition(name='c', type_='float'),
        ],
        returns=None
    )


def test_default_args():
    assert parse_function_definition(
        'def Foo(a: int, b: int=2, c: int=3) -> None: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[
            ArgumentDefinition(name='a', type_='int'),
            ArgumentDefinition(name='b', type_='int', has_default=True, default=2),
            ArgumentDefinition(name='c', type_='int', has_default=True, default=3),
        ],
        returns=None
    )


def test_returns_outer_iterator():
    assert parse_function_definition(
        'def Foo() -> Iterator[Tuple]: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=ReturnValueDefinition(
            outer_format=ReturnValueOuterFormat.ITERATOR,
            inner_format=ReturnValueInnerFormat.TUPLE
        )
    )


def test_returns_outer_list():
    assert parse_function_definition(
        'def Foo() -> List[Tuple]: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=ReturnValueDefinition(
            outer_format=ReturnValueOuterFormat.LIST,
            inner_format=ReturnValueInnerFormat.TUPLE
        )
    )


def test_returns_outer_single():
    assert parse_function_definition(
        'def Foo() -> Single[Tuple]: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=ReturnValueDefinition(
            outer_format=ReturnValueOuterFormat.SINGLE,
            inner_format=ReturnValueInnerFormat.TUPLE
        )
    )


def test_returns_inner_tuple():
    assert parse_function_definition(
        'def Foo() -> Single[Tuple]: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=ReturnValueDefinition(
            outer_format=ReturnValueOuterFormat.SINGLE,
            inner_format=ReturnValueInnerFormat.TUPLE
        )
    )


def test_returns_inner_dict():
    assert parse_function_definition(
        'def Foo() -> Single[Dict]: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=ReturnValueDefinition(
            outer_format=ReturnValueOuterFormat.SINGLE,
            inner_format=ReturnValueInnerFormat.DICT
        )
    )


def test_returns_inner_list():
    assert parse_function_definition(
        'def Foo() -> Single[List]: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=ReturnValueDefinition(
            outer_format=ReturnValueOuterFormat.SINGLE,
            inner_format=ReturnValueInnerFormat.LIST
        )
    )


def test_returns_inner_custom():
    assert parse_function_definition(
        'def Foo() -> Single[MyType]: ...'
    ) == FunctionDefinition(
        name='Foo',
        args=[],
        returns=ReturnValueDefinition(
            outer_format=ReturnValueOuterFormat.SINGLE,
            inner_format='MyType'
        )
    )


def test_returns_invalid():
    with pytest.raises(TypeError):
        parse_function_definition('def Foo() -> BadType[List]: ...')



# Copyright (c) 2020 Dmitry Marakasov <amdmi3@amdmi3.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ast
from dataclasses import dataclass, field
from enum import Enum, unique
from typing import Any, List, Optional, Union


@dataclass
class ArgumentDefinition:
    name: str
    type_: Any = None
    has_default: bool = False
    default: Any = None


@unique
class ReturnValueOuterFormat(Enum):
    ITERATOR = 1
    LIST = 2
    SINGLE = 3
    # DICT = 4  # TODO


@unique
class ReturnValueInnerFormat(Enum):
    TUPLE = 1
    DICT = 2
    LIST = 3


@dataclass
class ReturnValueDefinition:
    outer_format: ReturnValueOuterFormat
    inner_format: Union[ReturnValueInnerFormat, str]


@dataclass
class FunctionDefinition:
    name: str
    args: List[ArgumentDefinition] = field(default_factory=list)
    returns: Optional[ReturnValueDefinition] = None


def parse_function_definition(source: str) -> FunctionDefinition:
    tree = ast.parse(source)

    assert(len(tree.body) == 1)
    assert(isinstance(tree.body[0], ast.FunctionDef))

    func = tree.body[0]

    func_def = FunctionDefinition(name=func.name)

    # parse arguments
    for arg in func.args.args:
        arg_def = ArgumentDefinition(
            name=arg.arg,
            type_=arg.annotation.id if arg.annotation else None  # type: ignore
        )

        func_def.args.append(arg_def)

    # parse default values
    first_default_idx = len(func.args.args) - len(func.args.defaults)

    for argn, default in enumerate(func.args.defaults, first_default_idx):
        assert(isinstance(default, ast.Constant))

        func_def.args[argn].has_default = True
        func_def.args[argn].default = default.value

    # parse return value
    returns = func.returns

    if isinstance(returns, ast.Constant) and returns.value is None:
        func_def.returns = None
    else:
        assert(isinstance(returns, ast.Subscript))

        if returns.value.id in ('List', 'list'):  # type: ignore
            outer_format = ReturnValueOuterFormat.LIST
        elif returns.value.id == 'Iterator':  # type: ignore
            outer_format = ReturnValueOuterFormat.ITERATOR
        elif returns.value.id == 'Single':  # type: ignore
            outer_format = ReturnValueOuterFormat.SINGLE
        else:
            raise TypeError(f'Unexpected return value type {returns.value.id}')  # type: ignore

        if returns.slice.id in ('Tuple', 'tuple'):  # type: ignore
            inner_format = ReturnValueInnerFormat.TUPLE
        elif returns.slice.id in ('Dict', 'dict'):  # type: ignore
            inner_format = ReturnValueInnerFormat.DICT
        elif returns.slice.id in ('List', 'list'):  # type: ignore
            inner_format = ReturnValueInnerFormat.LIST
        else:  # custom type
            inner_format = returns.slice.id  # type: ignore

        func_def.returns = ReturnValueDefinition(outer_format, inner_format)

    # check body
    assert(len(func.body) == 1)
    assert(isinstance(func.body[0], ast.Expr))
    assert(isinstance(func.body[0].value, ast.Constant))
    assert(func.body[0].value.value is Ellipsis)

    return func_def



import pytest

from aesqlapius.namespace import Namespace, inject_method


@pytest.fixture
def method():
    pass


@pytest.fixture
def root():
    return Namespace()


def test_inject_into_root(root, method):
    inject_method(root, ['method'], method)
    assert root.method is method


def test_inject_into_subns(root, method):
    inject_method(root, ['subns1', 'subns2', 'method'], method)
    assert root.subns1.subns2.method is method


def test_duplicate_method(root, method):
    inject_method(root, ['method'], method)
    with pytest.raises(ValueError):
        inject_method(root, ['method'], method)


def test_duplicate_method_ns(root, method):
    inject_method(root, ['subns'], method)
    with pytest.raises(ValueError):
        inject_method(root, ['subns', 'method'], method)



#!/usr/bin/env python3

from os import path

from setuptools import find_packages, setup


here = path.abspath(path.dirname(__file__))


def get_version():
    with open(path.join(here, 'aesqlapius', '__init__.py')) as source:
        for line in source:
            if line.startswith('__version__'):
                return line.strip().split(' = ')[-1].strip("'")

    raise RuntimeError('Cannot determine package version from package source')


def get_long_description():
    try:
        return open(path.join(here, 'README.md')).read()
    except:
        return None


setup(
    name='aesqlapius',
    version=get_version(),
    description='Manage SQL queries as a Python API',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Dmitry Marakasov',
    author_email='amdmi3@amdmi3.ru',
    url='https://github.com/AMDmi3/aesqlapius',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: SQL',
        'Topic :: Database',
    ],
    python_requires='>=3.6',
    packages=find_packages(),
)



# Copyright (c) 2020 Dmitry Marakasov <amdmi3@amdmi3.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import Any, Dict, Iterator, List, Tuple

from aesqlapius.function_def import ArgumentDefinition, FunctionDefinition


def _iter_args(func_def: FunctionDefinition, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Iterator[Tuple[ArgumentDefinition, Any]]:
    for narg, arg in enumerate(func_def.args):
        print(arg)
        value: Any = None
        if narg < len(args):
            if arg.name in kwargs:
                raise TypeError(f"{func_def.name} got multiple values for argument '{arg.name}'")
            value = args[narg]
        elif arg.name in kwargs:
            value = kwargs[arg.name]
        elif arg.has_default:
            value = arg.default
        else:
            raise TypeError(f"{func_def.name} missing required argument '{arg.name}'")

        yield arg, value


def prepare_args_as_dict(func_def: FunctionDefinition, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Dict[str, Any]:
    return {arg.name: value for arg, value in _iter_args(func_def, args, kwargs)}


def prepare_args_as_list(func_def, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> List[Any]:
    return [value for arg, value in _iter_args(func_def, args, kwargs)]



# Copyright (c) 2020 Dmitry Marakasov <amdmi3@amdmi3.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
from dataclasses import dataclass
from typing import Iterator, List, Optional, Tuple

from aesqlapius.query import Query, parse_query_file


@dataclass
class QueryDirEntry:
    namespace_path: List[str]
    filesystem_path: str


def _walk_dir_tree(path: str, extension: str, namespace_path: Optional[List[str]] = None) -> Iterator[QueryDirEntry]:
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_dir():
                yield from _walk_dir_tree(
                    os.path.join(path, entry.name),
                    extension,
                    (namespace_path or []) + [entry.name]
                )
            elif entry.is_file() and entry.name.endswith(extension):
                yield QueryDirEntry(
                    (namespace_path or []) + [entry.name.removesuffix(extension)],
                    os.path.join(path, entry.name)
                )


def iter_query_dir(path: str, extension: str) -> Iterator[Tuple[QueryDirEntry, List[Query]]]:
    for entry in _walk_dir_tree(path, extension):
        yield (entry, parse_query_file(entry.filesystem_path))



# Copyright (c) 2020 Dmitry Marakasov <amdmi3@amdmi3.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import functools
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
    overload
)

from aesqlapius.args import prepare_args_as_dict
from aesqlapius.function_def import ReturnValueOuterFormat
from aesqlapius.namespace import Namespace, inject_method
from aesqlapius.output import generate_row_processor
from aesqlapius.query import Query
from aesqlapius.querydir import iter_query_dir


def _generate_method(query: Query) -> Callable[..., Any]:
    func_def = query.func_def
    returns = func_def.returns

    if returns is None:
        def method_returning_none(db, *args, **kwargs) -> None:
            with db.cursor() as cur:
                cur.execute(query.text, prepare_args_as_dict(func_def, args, kwargs))

        return method_returning_none

    elif returns.outer_format == ReturnValueOuterFormat.ITERATOR:
        def method_returning_iterator(db, *args, **kwargs) -> Iterator[Any]:
            assert(returns is not None)  # mypy bug
            with db.cursor() as cur:
                cur.execute(query.text, prepare_args_as_dict(func_def, args, kwargs))
                names = [desc.name for desc in cur.description]
                process_row = generate_row_processor(returns.inner_format, names)
                yield from map(process_row, cur)

        return method_returning_iterator

    elif returns.outer_format == ReturnValueOuterFormat.LIST:
        def method_returning_list(db, *args, **kwargs) -> List[Any]:
            assert(returns is not None)  # mypy bug
            with db.cursor() as cur:
                cur.execute(query.text, prepare_args_as_dict(func_def, args, kwargs))
                names = [desc.name for desc in cur.description]
                process_row = generate_row_processor(returns.inner_format, names)
                return [process_row(row) for row in cur]

        return method_returning_list

    elif returns.outer_format == ReturnValueOuterFormat.SINGLE:
        def method_returning_single(db, *args, **kwargs) -> Any:
            assert(returns is not None)  # mypy bug
            with db.cursor() as cur:
                cur.execute(query.text, prepare_args_as_dict(func_def, args, kwargs))
                names = [desc.name for desc in cur.description]
                process_row = generate_row_processor(returns.inner_format, names)
                return process_row(cur.fetchone())

        return method_returning_single

    else:
        raise NotImplementedError(f"unsupported outer return type format '{returns.outer_format}'")  # pragma: no cover


def _maybe_bind(query: Query, db: Any) -> Callable[..., Any]:
    if db is None:
        return _generate_method(query)
    else:
        return functools.partial(_generate_method(query), db)


T = TypeVar('T')


@overload
def generate_api(
    path: str,
    db: Any = None,
    *,
    file_as_namespace: bool = False,
    extension: str = '.sql',
    namespace_root: str = '__init__',
) -> Namespace:
    ...  # pragma: no cover


@overload
def generate_api(
    path: str,
    db: Any = None,
    *,
    target: T,
    file_as_namespace: bool = False,
    extension: str = '.sql',
    namespace_root: str = '__init__',
) -> T:
    ...  # pragma: no cover


def generate_api(
    path: str,
    db: Any = None,
    *,
    target: Optional[T] = None,
    file_as_namespace: bool = False,
    extension: str = '.sql',
    namespace_root: str = '__init__',
) -> Union[T, Namespace]:
    ns: Union[T, Namespace]
    if target is None:
        ns = Namespace()
    else:
        ns = target

    for entry, queries in iter_query_dir(path, extension):
        if file_as_namespace and entry.namespace_path[-1] != namespace_root:
            namespace_path = entry.namespace_path
        else:
            namespace_path = entry.namespace_path[:-1]

        for query in queries:
            inject_method(
                ns,
                namespace_path + [query.func_def.name],
                _maybe_bind(query, db)
            )

    return ns
